/*
 * cbm/buffer.c
 *
 * Modified from msdos/buffer.c
 *
 * 95-03-25  Dan Fandrich <dan@fch.wimsey.bc.ca>
 * 97-03-02  dan  Updated for kernel 2.0.29
 */

#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/fs.h>
#include <linux/cbm_fs.h>

/* This version number probably isn't accurate */
#if LINUX_VERSION_CODE < 0x010399       /* kernel ver. < 1.3.99 */
static inline int buffer_uptodate(struct buffer_head * bh)
{
	return bh->b_uptodate;
}

static inline void mark_buffer_uptodate(struct buffer_head * bh, int on)
{
	bh->b_uptodate = on;
}
#endif


/* #Specification: cbm / strategy / special device / dummy blocks
	The Commodore 1581 disk format uses 512 byte physical sectors
	partitioned into 2 256 byte logical sectors.  Since these
	sectors can hold data for 2 different files, this means that
	the hardware sector must be read, patched and written for almost
	all writes.

	Needless to say that it kills write performance
	on all OSes.

	Internally the Linux cbm fs is using 256 byte
	logical sectors. When accessing such a device, we
	allocate dummy buffer cache blocks, that we stuff
	with the information of a real one (512 bytes large).

	This strategy is used to hide this difference to
	the core of the cbm fs. The slowdown is not
	hidden though!

	Because Linux doesn't support 256 byte buffers, we must use
	these routines for all reads as well as writes in order to
	have the 256->512 byte sector number conversion done
	transparently.
*/

/* Do a bread where block is a 256-byte logical CBM sector */
struct buffer_head *cbm_bread (
	struct super_block *sb,
	int block)
{
	struct buffer_head *ret = NULL;
	struct buffer_head *real = bread (sb->s_dev,block>>1,CBM_BLOCK_SIZE);
	if (real != NULL){
		ret = (struct buffer_head *)kmalloc (sizeof(struct buffer_head)
			,GFP_KERNEL);
		if (ret != NULL){
			/*
				The memset is there only to catch errors. The cbm
				fs only uses b_data, and not the other dozen or so
				fields, so we can save time by not copying them.
			*/
			memset (ret,0,sizeof(*ret));
			ret->b_data = real->b_data;
			if (block & 1) ret->b_data += CBM_LBLOCK_SIZE;
			ret->b_next = real;

		}else{
			/*
				Error allocating memory for copy of buffer, so release
				original buffer and exit
			 */
			brelse (real);
		}
	}
	return ret;
}

/* This routine can probably be eliminated */
struct buffer_head *cbm_getblk (
	struct super_block *sb,
	int block)
{
	struct buffer_head *ret = NULL;
	/* #Specification: msdos / special device / writing
		A write is always preceded by a read of the complete block
		(large hardware sector size). This defeat write performance.
		There is a possibility to optimize this when writing large
		chunk by making sure we are filling large block. Volunteer ?
	*/
	ret = cbm_bread (sb,block);
	return ret;
}

void cbm_brelse ( struct buffer_head *bh )
{
	if (bh != NULL){
		/* Release the real buffer */
		brelse (bh->b_next);
		/* We can free the dummy because a new one is allocated at
			each cbm_getblk() and cbm_bread().
		*/
		kfree (bh);
	}
}
	
/* The msdos fs analogue isn't used, so we probably won't either.
Mark the buffer as dirty (or clean) */
#if MAY_NEED_FOR_FUTURE_WRITE_SUPPORT
void cbm_mark_buffer_dirty (
	struct super_block *sb,
	struct buffer_head *bh,
	int dirty_val)
{
	bh = bh->b_next;	/* point to real buffer */
	mark_buffer_dirty (bh,dirty_val);
}
#endif

void cbm_set_uptodate (
	struct super_block *sb,
	struct buffer_head *bh,
	int val)
{
	bh = bh->b_next;	/* point to real buffer */
	mark_buffer_uptodate(bh, val);
}

int cbm_is_uptodate (
	struct super_block *sb,
	struct buffer_head *bh)
{
	bh = bh->b_next;	/* point to real buffer */
	return buffer_uptodate(bh);
}

void cbm_ll_rw_block (
	struct super_block *sb,
	int opr,
	int nbreq,
	struct buffer_head *bh[CBM_PREFETCH])
{
	struct buffer_head *tmp[CBM_PREFETCH];
	int i;
	for (i=0; i<nbreq; i++){
		tmp[i] = bh[i]->b_next;
	}
	ll_rw_block(opr,nbreq,tmp);
}

