/*
 *  linux/fs/cbm/dir.c
 *
 *  Written 1995 by Daniel Fandrich <dan@fch.wimsey.bc.ca>
 *  Modified from MS-DOS filesystem written 1992,1993 by Werner Almesberger
 *
 *  CBM directory handling functions
 */

#include <asm/segment.h>

#include <linux/cbm_fs.h>
#include <linux/cbm_fs_i.h>
#include <linux/cbm_fs_sb.h>
#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/stat.h>
#include <linux/string.h>
#include <linux/kernel.h>	/* for printk */

#include "cbmbuffer.h"

/* Return the offset of the last (name) entry in dirent */
#define NAME_OFFSET(de) ((int) ((de)->d_name - (char *) (de)))
/* Rounds up a number to the nearest multiple of 4 */
#define ROUND_UP(x) (((x)+3) & ~3)

#define PRINTK(X) /*printk X*/	/* debugging */

static int cbm_dir_read(struct inode * inode,struct file * filp, char * buf,int count)
{
	return -EISDIR;
}

static struct file_operations cbm_dir_operations = {
	NULL,			/* lseek - default */
	cbm_dir_read,	/* read */
	NULL,			/* write - bad */
	cbm_readdir,	/* readdir */
	NULL,			/* select - default */
	cbm_ioctl,		/* ioctl */
	NULL,			/* mmap */
	NULL,			/* no special open code */
	NULL,			/* no special release code */
	file_fsync		/* fsync */
};

/* What are the implications of supporting the cbm_bmap call? */
struct inode_operations cbm_dir_inode_operations = {
	&cbm_dir_operations,	/* default directory file-ops */
	NULL /*cbm_create*/,		/* create */
	cbm_lookup,		/* lookup */
	NULL,			/* link */
	NULL /*cbm_unlink*/,		/* unlink */
	NULL,			/* symlink */
	NULL /*cbm_mkdir*/,		/* mkdir */
	NULL /*cbm_rmdir*/,		/* rmdir */
	NULL,			/* mknod */
	NULL /*cbm_rename*/,		/* rename */
	NULL,			/* readlink */
	NULL,			/* follow_link */
	cbm_bmap,		/* bmap */
	NULL,			/* truncate */
	NULL			/* permission */
};

/* Formats a UNIX file name given a CBM name. */
/* Input: nous, name */
/* Returns: length of converted file name (0 on error) */
static int fs_format_name(int nous,const unsigned char name[],
							unsigned char res[])
{
	int i,last;
	unsigned char c;

	/* Convert entire file name from PETSCII to ASCII */
	for (i = last = 0; i < CBM_NAME; i++) {
		if (!(c = name[i]) || (c == CBM_END_NAME)) break;
		if (c >= 'A' && c <= 'Z') c += 0x20;
		else if (c >= 'a' && c <= 'z') c -= 0x20;
		else if (c == '/') c = '\\';
		c &= 0x7f;		/* Upper case characters have bit 7 set */

		/* convert spaces to underscores unless requested otherwise */
		if ((c == ' ') && !nous)
			c = '_';
		/*if (c & 0x7f != ' ')*/	/* kill trailing spaces */
			last = i+1;
		res[i] = c;
	}
	res[last] = '\0';
	PRINTK (("readdir name \"%s\"\n", res));

	return last;
}

/* Fill directory structure */
#if LINUX_VERSION_CODE >= 0x010300	/* kernel ver. >= 1.3.0 */

#define filldir kfilldir	/* use kernel's filldir, passed to cbm_readdir */

#else

/* Fill directory structure in user space */
/* Can only fill a single entry */
/* name need not be null-terminated */
int filldir(struct dirent *dirent, const char *name, int len, int offs, ino_t ino)
{
	memcpy_tofs(dirent->d_name,name,len);
	put_fs_byte(0,dirent->d_name+len);		/* null-terminate string */
	put_fs_word(len,&dirent->d_reclen);
	put_fs_long(ino,&dirent->d_ino);
	put_fs_long(ROUND_UP(NAME_OFFSET(dirent) + len + 1),&dirent->d_off);

	return 0;
}
#endif

/* Read the contents of a CBM directory `file' */
/* For kernel < 1.3, returns the sizeof dirent including the name or -error */
/* For kernel >= 1.3, return 0 or -error */
#if LINUX_VERSION_CODE >= 0x010300	/* kernel ver. >= 1.3.0 */
int cbm_readdir(struct inode *inode, struct file *filp, void *dirent,
				filldir_t kfilldir)
#else
/* dirent in in user space */
int cbm_readdir(struct inode *inode, struct file *filp, struct dirent *dirent,
				int count)
#endif
{
	int ino,nlen;
	struct buffer_head *bh;
	struct cbm_dir_entry *de;
	unsigned long oldpos = filp->f_pos;

	PRINTK (("readdir enter %ld\n",inode->i_ino));

	/* Is inode actually a directory?  VFS should only give us directories. */
	if (!inode || !S_ISDIR(inode->i_mode)) return -EBADF;

	PRINTK (("readdir after !S_ISDIR\n"));

	/* Fake "." and ".." directories */

	/* f_pos==0 means "." */
	/* f_pos==1 means ".." */
	/* f_pos==2 means the first real entry */
	if (oldpos == 2)
		filp->f_pos = 0;	/* go get first real entry */

	else if (oldpos < 2) {				/* "." and ".." virtual entries */

		if (filldir(dirent, "..", oldpos+1, oldpos,
					oldpos == 0 ? inode->i_ino : CBM_I(inode)->i_parent) < 0)
			return 0;

		filp->f_pos++;		/* increment f_pos for the next entry */

#if LINUX_VERSION_CODE >= 0x010300	/* kernel ver. >= 1.3.0 */
		return 0;
#else
		return ROUND_UP(NAME_OFFSET(dirent) + oldpos + 2);	/* exit */
#endif
	}

	PRINTK (("readdir want real entry\n"));

	/* Sanity check: is request at the beginning of a full directory entry? */
	if (((int) filp->f_pos % CBM_DATA_SIZE) % sizeof(struct cbm_dir_entry))
		return -ENOENT;

	PRINTK (("readdir loop\n"));
	/* Loop through the directory, searching for a good directory entry */
	bh = NULL;
	while ((ino = cbm_get_entry(inode,&filp->f_pos,&bh,&de)) >= 0) {
		if (!IS_FREE(de->type)) {
			unsigned char bufname[CBM_NAME];

			/* convert name to CBM format */
			nlen = fs_format_name(CBM_SB(inode->i_sb)->nous, de->name, bufname);

			/* make sure converted file name isn't empty */
			if (nlen > 0) {

				if (filldir(dirent, bufname, nlen, oldpos, ino) < 0) {
					filp->f_pos = oldpos;
					break;		/* exit scan loop */
				}

				cbm_brelse(bh);

#if LINUX_VERSION_CODE >= 0x010300	/* kernel ver. >= 1.3.0 */
				return 0;
#else
				return ROUND_UP(NAME_OFFSET(dirent) + nlen + 1);
#endif
			}
		}
		oldpos = filp->f_pos;
	}
	if (bh) cbm_brelse(bh);
	return 0;
}
