/*
 *  linux/fs/cbm/file.c
 *
 *  Written 1995 by Dan Fandrich <dan@fch.wimsey.bc.ca>
 *  Modified from MS-DOS filesystem written 1992,1993 by Werner Almesberger
 *
 *  CBM regular file handling primitives
 */

#include <asm/segment.h>
#include <asm/system.h>

#include <linux/cbm_fs.h>
#include <linux/cbm_fs_sb.h>
#include <linux/cbm_fs_i.h>
#include <linux/sched.h>
#include <linux/locks.h>
#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/fcntl.h>
#include <linux/stat.h>
#include <linux/string.h>

#include "cbmbuffer.h"

#define MIN(a,b) (((a) < (b)) ? (a) : (b))
#define MAX(a,b) (((a) > (b)) ? (a) : (b))

#define PRINTK(x) /*printk x*/		/* Debugging */

/* #Specification: msdos / special devices / mmap	
	Mmapping does work because a special mmap is provide in that case.
	Note that it is much less efficient than the generic_mmap normally
	used since it allocate extra buffer. generic_mmap is used for
	normal device (512 bytes hardware sectors).
*/
static struct file_operations cbm_file_operations = {
	NULL,			/* lseek - default */
	cbm_file_read,	/* read */
	NULL /*cbm_file_write*/,	/* write */
	NULL,			/* readdir - bad */
	NULL,			/* select - default */
	cbm_ioctl,			/* ioctl */
	NULL /*cbm_mmap*/,		/* mmap */
	NULL,			/* no special open is needed */
	NULL,			/* release */
	file_fsync		/* fsync */
};

/* #Specification: msdos / special devices / swap file
	Swap file can't work on special devices with a large sector
	size (1024 bytes hard sector). Those devices have a weird
	MSDOS filesystem layout. Generally a single hardware sector
	may contain 2 unrelated logical sector. This mean that there is
	no easy way to do a mapping between disk sector of a file and virtual
	memory. So swap file is difficult (not available right now)
	on those devices. Off course, Ext2 does not have this problem.
*/
struct inode_operations cbm_file_inode_operations = {
	&cbm_file_operations,	/* default file operations */
	NULL,			/* create */
	NULL,			/* lookup */
	NULL,			/* link */
	NULL,			/* unlink */
	NULL,			/* symlink */
	NULL,			/* mkdir */
	NULL,			/* rmdir */
	NULL,			/* mknod */
	NULL,			/* rename */
	NULL,			/* readlink */
	NULL,			/* follow_link */
	NULL,			/* bmap */  /* NOTE: should this be cbm_bmap?
								  Can we have a bmap because of 254 byte sec? */
	NULL /*cbm_truncate*/,		/* truncate */
	NULL,			/* permission */
	NULL			/* smap */
};

/* sector prefetch list */
struct cbm_pre {
	int file_sector;		/* Next sector to read in the prefetch table */
							/* This is relative to the file, not the disk */
	struct buffer_head *bhlist[CBM_PREFETCH];		/* All buffers needed */
	int nblist;				/* Number of buffers in bhlist */
	int nolist;				/* index in bhlist */
};

/*
	Order the prefetch of nb more sectors, help in *pre.
*/
static void cbm_prefetch (
	struct inode *inode,
	struct cbm_pre *pre,
	int nb)		/* How many must be prefetched at once */
{
	struct super_block *sb = inode->i_sb;
	struct buffer_head *bhreq[CBM_PREFETCH];	/* Buffers not already read */
	int nbreq=0;			/* Number of buffers in bhreq */
	int i;

	/* For nb sectors, map the logical sectors onto the disk */
	for (i=0; i<nb; i++){
		int sector = cbm_smap(inode,pre->file_sector);
		if (sector >= 0){
			struct buffer_head *bh;

			PRINTK (("prefetch: fsector2 %d -> %d\n",pre->file_sector-1,sector));
			pre->file_sector++;

			/* read the sector from disk */
			bh = cbm_getblk(inode->i_sb,sector);
			if (bh == NULL)	break;		/* error reading sector */

			/* put the buffer into the list and increment the count */
			pre->bhlist[pre->nblist++] = bh;
			if (!cbm_is_uptodate(sb,bh)) bhreq[nbreq++] = bh;

		} else			/* error in mapping */
			break;
	}

	if (nbreq > 0) cbm_ll_rw_block (sb,READ,nbreq,bhreq);
	
	/* we fetched all the buffers, so clear them */
	for (i=pre->nblist; i<CBM_PREFETCH; i++) pre->bhlist[i] = NULL;
}

/*
	Read a file into user space
*/
int cbm_file_read(
	struct inode *inode,
	struct file *filp,
	char *buf,
	int count)
{
	struct super_block *sb = inode->i_sb;
	char *start = buf;
	char *end   = buf + count;
	int i;
	int left_in_file,i_size;
	struct cbm_pre pre;
		
	if (!inode) {
		printk("cbm_file_read: inode = NULL\n");
		return -EINVAL;
	}

	/* Make sure it's a regular file */
	if (!S_ISREG(inode->i_mode)) {
		printk("cbm_file_read: mode = %07o\n",inode->i_mode);
		return -EINVAL;
	}

	/* Make sure file position isn't past the end of file */
	if (filp->f_pos >= inode->i_size || count <= 0) return 0;

	/*
		Tell the buffer cache which block we expect to read in advance
		Since we are limited with the stack, we prefetch only CBM_PREFETCH
		because we have to keep the result into the local
		arrays pre.bhlist and bhreq.
		
		Each time we process one block in bhlist, we replace
		it by a new prefetch block if needed.
	*/
	PRINTK (("#### ino %ld pos %ld size %ld count %d\n",
				inode->i_ino,(long)filp->f_pos,inode->i_size,count));
	{
		/*
			We must prefetch complete block, so we must
			take in account the offset in the first block.
		*/
		int count_max = ((long) filp->f_pos % CBM_DATA_SIZE) + count;
		/* How many block to read all at once */
		int   to_reada = count_max / CBM_DATA_SIZE;
		pre.file_sector = (long) filp->f_pos / CBM_DATA_SIZE;

		/* Does count_max spill over into another sector? */
		if (count_max % CBM_DATA_SIZE) to_reada++;

		if (filp->f_reada){
			int ahead = read_ahead[MAJOR(inode->i_dev)];
			PRINTK (("to_reada %d ahead %d\n",to_reada,ahead));
			if (ahead == 0) ahead = 8;		/* Why 8?--DF */
			to_reada += ahead;
		}
		to_reada = MIN(to_reada,CBM_PREFETCH);
		pre.nblist = 0;
		cbm_prefetch (inode,&pre,to_reada);
	}

	PRINTK (("count %d ahead %d nblist %d\n",
		count,read_ahead[MAJOR(inode->i_dev)],pre.nblist));

	pre.nolist = 0;			/* index into the buffer array */

	/* Read all blocks in the file, one at a time */
	while ((left_in_file = (i_size = inode->i_size) - filp->f_pos) > 0
			&& buf < end){
		struct buffer_head *bh = pre.bhlist[pre.nolist];
		char *data;
		int size,offset;

		/* is buffer ready? */
		if (bh == NULL) break;

		/* we have buffer--clear from list & increment the index */
		pre.bhlist[pre.nolist] = NULL;
		pre.nolist++;

		/* When we're halfway through the prefetch buffer, move the last
		   half of the contents down to make room for more */
		if (pre.nolist == CBM_PREFETCH/2){
			memcpy (pre.bhlist, pre.bhlist+CBM_PREFETCH/2,
					(CBM_PREFETCH/2)*sizeof(pre.bhlist[0]));
			pre.nblist -= CBM_PREFETCH/2;
			cbm_prefetch (inode, &pre, CBM_PREFETCH/2);
			pre.nolist = 0;		/* start the index from the beginning again */
		}
		PRINTK (("file_read pos %ld nblist %d %d %d\n",
			(long)filp->f_pos,pre.nblist,pre.nolist,count));
		wait_on_buffer(bh);
		if (!cbm_is_uptodate(sb,bh)){
			/* read error  ? */
			cbm_brelse (bh);
			break;
		}

		offset = (long) filp->f_pos % CBM_DATA_SIZE;	/* offset within sector */
		size = MIN(CBM_DATA_SIZE-offset,left_in_file);	/* bytes in sector */
		data = ((struct cbm_data_sector *) bh->b_data)->data + offset;

		/* Make sure we're not trying to read past the end of file.
		   This can happen in fastsize mode, where file sizes are rounded up.
		   If so, set cached inode size to real file size for future use. */
		/* Use the copy of inode->i_size in i_size as the base to reduce the
		   size to avoid a race condition if two processes simultaneously
		   decide to adjust the size in the same file's inode */
		if (CBM_SB(sb)->fastsize &&
			(((struct cbm_data_sector *) bh->b_data)->t == 0) &&
			(size > (int)((struct cbm_data_sector *) bh->b_data)->s-1-offset)) {

			/* This works because the estimated size is never going to be more
			   than 253 bytes larger than the actual size */
			inode->i_size = i_size -= CBM_DATA_SIZE -
								((struct cbm_data_sector *) bh->b_data)->s + 1;

			/* reducing the inode size might actually make the request
			   fail due to reading past the real end of file */
			size = MAX(((struct cbm_data_sector *) bh->b_data)->s-1-offset, 0);
			PRINTK(("file_read: adjusting inode size %ld\n",(long)inode->i_size));
		}

		/* Binary or text file? */
		if (!CBM_I(inode)->i_text) {

			/* Binary file--copy buffer contents exactly */
			size = MIN(size,end-buf);
			memcpy_tofs(buf,data,size);
			buf += size;
			filp->f_pos += size;
		}else{

			/* Do PETSCII->ASCII conversion on a text file */
			for (; size && buf < end; size--) {
				char ch = *data++;
				filp->f_pos++;
				ch = 0x7f & ((ch >= 'a' && ch <= 'z') ? ch-32 :
					 ((ch >= 'A' && ch <= 'Z') ? ch+32 :
					 ((ch == '\r') ? '\n' : ch)));
				put_fs_byte(ch,buf++);
			}
		}
		cbm_brelse(bh);

	} /* while */
	PRINTK (("--- %d -> %d\n",count,(int)(buf-start)));

	/* release all blocks in the prefetch list */
	for (i=0; i<pre.nblist; i++) cbm_brelse (pre.bhlist[i]);

	/* Error if nothing was copied into the destination buffer */
	if (start == buf) return -EIO;

	/* Update the file's access time if the inode is writable (this is pretty
	   meaningless since CBM can't store file dates) */
	if (!IS_RDONLY(inode)) inode->i_atime = CURRENT_TIME;

	filp->f_reada = 1;	/* Will be reset if a lseek is done */
	return buf-start;	/* Number of bytes put into destination buffer */
}

#if MAY_NEED_FOR_FUTURE_WRITE_SUPPORT
/*
	Write to a file either from user space
*/
int msdos_file_write(
	struct inode *inode,
	struct file *filp,
	char *buf,
	int count)
{
	struct super_block *sb = inode->i_sb;
	int sector,offset,size,left,written;
	int error,carry;
	char *start,*to,ch;
	struct buffer_head *bh;
	int binary_mode = MSDOS_I(inode)->i_binary;

	if (!inode) {
		printk("msdos_file_write: inode = NULL\n");
		return -EINVAL;
	}
	/* S_ISLNK allows for UMSDOS. Should never happen for normal MSDOS */
	if (!S_ISREG(inode->i_mode) && !S_ISLNK(inode->i_mode)) {
		printk("msdos_file_write: mode = %07o\n",inode->i_mode);
		return -EINVAL;
	}
/*
 * ok, append may not work when many processes are writing at the same time
 * but so what. That way leads to madness anyway.
 */
	if (filp->f_flags & O_APPEND) filp->f_pos = inode->i_size;
	if (count <= 0) return 0;
	error = carry = 0;
	for (start = buf; count || carry; count -= size) {
		while (!(sector = msdos_smap(inode,filp->f_pos >> SECTOR_BITS)))
			if ((error = msdos_add_cluster(inode)) < 0) break;
		if (error) {
			msdos_truncate(inode);
			break;
		}
		offset = filp->f_pos & (SECTOR_SIZE-1);
		size = MIN(SECTOR_SIZE-offset,MAX(carry,count));
		if (binary_mode
			&& offset == 0
			&& (size == SECTOR_SIZE
				|| filp->f_pos + size >= inode->i_size)){
			/* No need to read the block first since we will */
			/* completely overwrite it */
			/* or at least write past the end of file */
			if (!(bh = getblk(inode->i_dev,sector,SECTOR_SIZE))){
				error = -EIO;
				break;
			}
		}else if (!(bh = bread(inode->i_dev,sector,SECTOR_SIZE))) {
			error = -EIO;
			break;
		}
		if (binary_mode) {
			memcpy_fromfs(bh->b_data+offset,buf,written = size);
			buf += size;
		}
		else {
			written = left = SECTOR_SIZE-offset;
			to = (char *) bh->b_data+(filp->f_pos & (SECTOR_SIZE-1));
			if (carry) {
				*to++ = '\n';
				left--;
				carry = 0;
			}
			for (size = 0; size < count && left; size++) {
				if ((ch = get_fs_byte(buf++)) == '\n') {
					*to++ = '\r';
					left--;
				}
				if (!left) carry = 1;
				else {
					*to++ = ch;
					left--;
				}
			}
			written -= left;
		}
		filp->f_pos += written;
		if (filp->f_pos > inode->i_size) {
			inode->i_size = filp->f_pos;
			inode->i_dirt = 1;
		}
		msdos_set_uptodate(sb,bh,1);
		mark_buffer_dirty(bh, 0);
		brelse(bh);
	}
	if (start == buf)
		return error;
	inode->i_mtime = inode->i_ctime = CURRENT_TIME;
	MSDOS_I(inode)->i_attrs |= ATTR_ARCH;
	inode->i_dirt = 1;
	return buf-start;
}

void msdos_truncate(struct inode *inode)
{
	int cluster;

	cluster = SECTOR_SIZE*MSDOS_SB(inode->i_sb)->cluster_size;
	(void) fat_free(inode,(inode->i_size+(cluster-1))/cluster);
	MSDOS_I(inode)->i_attrs |= ATTR_ARCH;
	inode->i_dirt = 1;
}
#endif
