/*
 *  linux/fs/cbm/inode.c
 *
 *  Written 1995 by Dan Fandrich <dan@fch.wimsey.bc.ca>
 *
 * 95-03-25  converted from msdos/inode.c
 *
 */

/* REMEMBER: append mode probably cannot work with the fastsize option */

#include <linux/cbm_fs.h>
#include <linux/cbm_fs_sb.h>
#include <linux/cbm_fs_i.h>

#include <asm/segment.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/ctype.h>
#include <linux/major.h>
#include <linux/blkdev.h>
#include <linux/fs.h>
#include <linux/stat.h>
#include <linux/locks.h>

#include "cbmbuffer.h"

#ifdef MODULE
#include <linux/module.h>
#else

/* Only tested as a module: the rest of the kernel stuff isn't here */
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif

#define VERSION "0.3"	/* filesystem version number */

#define DEBUG 0 /*1*/

/* No writing for cbm yet */
#if MAY_NEED_FOR_FUTURE_WRITE_SUPPORT
void cbm_put_inode(struct inode *inode)
{
	struct super_block *sb;

	if (inode->i_nlink) {
		if (CBM_I(inode)->i_busy) cache_inval_inode(inode);
		return;
	}
	inode->i_size = 0;
	cbm_truncate(inode);
	sb = inode->i_sb;
	clear_inode(inode);
}
#endif


void cbm_put_super(struct super_block *sb)
{
	/*
	 * Invalidate our cache of file locations for this device
	 */

	cache_inval_dev(sb->s_dev);
	set_blocksize (sb->s_dev,BLOCK_SIZE);	/* What is this for?-DF */
	lock_super(sb);
	sb->s_dev = 0;
	unlock_super(sb);
	MOD_DEC_USE_COUNT;	/* we're done with this filesystem */
	return;
}

static struct super_operations cbm_sops = { 
	cbm_read_inode,
	cbm_notify_change,
	NULL 	/*msdos_write_inode*/,
	NULL 	/*cbm_put_inode*/,
	cbm_put_super,
	NULL, 	/* added in 0.96c */
	cbm_statfs,
	NULL
};

/*
 * Parse the mount options
 */

static int parse_options(char *options,char *check,char *conversion,int *fstype,
    uid_t *uid,gid_t *gid,int *umask,int *debug,int *quiet,int *fastsize,int *nous)
{
	char *this_char,*value;

	*check = 'n';
	*conversion = 'b';
	*uid = current->uid;
	*gid = current->gid;
	*umask = current->fs->umask;
	*fstype = *debug = *quiet = *fastsize = *nous = 0;
	if (!options) return 1;

	for (this_char = strtok(options,","); this_char; this_char = strtok(NULL,",")) {
		if ((value = strchr(this_char,'=')) != NULL)
			*value++ = 0;

		/*
		 * File name checking strictness
		 */

		if (!strcmp(this_char,"check") && value) {
			if (value[0] && !value[1] && strchr("rns",*value))
				*check = *value;
			/* else if (!strcmp(value,"relaxed")) *check = 'r'; */
			else if (!strcmp(value,"normal")) *check = 'n';
			else if (!strcmp(value,"strict")) *check = 's';
			else return 0;
		}

		/*
		 * PETSCII-ASCII conversion
		 */

		else if (!strcmp(this_char,"conv") && value) {
			if (value[0] && !value[1] && strchr("bta",*value))
				*conversion = *value;
			else if (!strcmp(value,"binary")) *conversion = 'b';
			else if (!strcmp(value,"text")) *conversion = 't';
			else if (!strcmp(value,"auto")) *conversion = 'a';
			else return 0;
		}

		/*
		 * Commodore disk type
		 */

		else if (!strcmp(this_char,"type")) {
			if (!value || !*value)
				return 0;
			*fstype = simple_strtoul(value,&value,0);
			if (*value || (*fstype != 1541 && *fstype != 1581))
				return 0;
		}

		/*
		 * User ID for files
		 */

		else if (!strcmp(this_char,"uid")) {
			if (!value || !*value)
				return 0;
			*uid = simple_strtoul(value,&value,0);
			if (*value)
				return 0;
		}

		/*
		 * Group ID for files
		 */

		else if (!strcmp(this_char,"gid")) {
			if (!value || !*value)
				return 0;
			*gid = simple_strtoul(value,&value,0);
			if (*value)
				return 0;
		}

		/*
		 * umask for files
		 */

		else if (!strcmp(this_char,"umask")) {
			if (!value || !*value)
				return 0;
			*umask = simple_strtoul(value,&value,8);
			if (*value)
				return 0;
		}

		/*
		 * Display debug information on mounting
		 */

		else if (!strcmp(this_char,"debug")) {
			if (value) return 0;
			*debug = 1;
		}

		/*
		 * Don't return errors when changing file attributes
		 */

		else if (!strcmp(this_char,"quiet")) {
			if (value) return 0;
			*quiet = 1;
		}

		/*
		 * Approximate file sizes
		 */

		else if (!strcmp(this_char,"fastsize")) {
			if (value) return 0;
			*fastsize = 1;
		}

		/*
		 * Show spaces in file names
		 */

		else if (!strcmp(this_char,"nous")) {
			if (value) return 0;
			*nous = 1;
		}
		else return 0;
	}
	return 1;
}


/*
 * Read the super block of Commodore 1541/1581 FS.
 * This is the first function called after mounting the filesystem
 */

struct super_block *cbm_read_super(struct super_block *sb,void *data,
				     int silent)
{
	struct buffer_head *bh;
	struct cbm1541_header_sector *b1541;
	struct cbm1581_header_sector *b1581;
	int fstype,debug,error,quiet,fastsize,nous;
	char check,conversion;
	uid_t uid;
	gid_t gid;
	int umask;
	int blksize = 512;

	MOD_INC_USE_COUNT;
	if (hardsect_size[MAJOR(sb->s_dev)] != NULL){
		blksize = hardsect_size[MAJOR(sb->s_dev)][MINOR(sb->s_dev)];
	}

	/*
	 * Parse mount options
	 */
	if (!parse_options((char *) data,&check,&conversion,&fstype,&uid,&gid,
		&umask,&debug,&quiet,&fastsize,&nous)
		|| (blksize != 512)) {
		/* Cannot handle this sector size--abort */
		sb->s_dev = 0;
		MOD_DEC_USE_COUNT;
		return NULL;
	}

	cache_init();
	lock_super(sb);
	sb->s_blocksize = CBM_BLOCK_SIZE; /* CBM_LBLOCK_SIZE probably isn't allowed */
	set_blocksize(sb->s_dev, CBM_BLOCK_SIZE);

	/*
	 * See if device contains a 1541 filesystem, or if a filesystem type is
	 * specified, make sure that the disk is readable
	 */

	if ((bh = cbm_bread(sb, CBM1541_SUPER_LOC)) == NULL) {
		unlock_super(sb);
		sb->s_dev = 0;
		printk("CBM-fs: unable to read 1541 superblock\n");
		MOD_DEC_USE_COUNT;
		return NULL;
	}

	b1541 = (struct cbm1541_header_sector *) bh->b_data;

	/*
	 * Try to determine if this is a 1541 disk or not
	 */

	error = !is_1541_header(b1541) || (b1541->t != CBM1541_DIR_TRACK);
										/* directory should be on track 18 */

	if ((!error && !fstype) || fstype == 1541) {
		error = 0;		/* make sure we don't test for a 1581 filesystem later */

		/* 1541 disk */
		CBM_SB(sb)->fs_type = 1541;
		CBM_SB(sb)->super_start = CBM1541_SUPER_LOC;
		CBM_SB(sb)->super_length = CBM1541_SUPER_LEN;
		CBM_SB(sb)->blocks = CBM1541_BLOCKS;
		CBM_SB(sb)->root_start = cbm1541_block(b1541->t, b1541->s);
			/* should always be t18 s1 */
		CBM_SB(sb)->root_high = cbm1541_block(b1541->t, 18);
			/* last reserved directory sector--t18 s18 */

#if DEBUG
		printk("[%d? BAM @%ld: %d %d %2x %2x]\n",
			CBM_SB(sb)->fs_type, CBM_SB(sb)->super_start,
			(int) b1541->t, (int) b1541->s,
			(int) b1541->disk_format, (int) b1541->flag);
#endif
		cbm_brelse(bh);		/* finished with block */

	} else {

		cbm_brelse(bh);		/* finished with 1541 block */

		/*
		 * See if device contains a 1581 filesystem
		 */
		if ((bh = cbm_bread(sb, CBM1581_SUPER_LOC)) == NULL) {
			unlock_super(sb);
			sb->s_dev = 0;
			printk("CBM-fs: unable to read 1581 superblock\n");
			MOD_DEC_USE_COUNT;
			return NULL;
		}
		b1581 = (struct cbm1581_header_sector *) bh->b_data;
		error = !is_1581_header(b1581) || (b1581->t != CBM1581_DIR_TRACK);
								/* root directory should be on track 40 */

		if ((!error && !fstype) || fstype == 1581) {
			error = 0;		/* either there is no error, or user says no error */

			/* 1581 disk */
			CBM_SB(sb)->fs_type = 1581;
			CBM_SB(sb)->super_start = CBM1581_SUPER_LOC;
			CBM_SB(sb)->super_length = CBM1581_SUPER_LEN;
			CBM_SB(sb)->blocks = CBM1581_BLOCKS;
			CBM_SB(sb)->root_start = cbm1581_block(b1581->t, b1581->s);
				/* should always be t40 s3 */
			CBM_SB(sb)->root_high = cbm1581_block(b1581->t, 39);
				/* last reserved directory sector--t40 s39 */
		}

#if DEBUG
		printk("[%d? BAM @%ld: %d %d %2x %2x]\n",
			CBM_SB(sb)->fs_type, CBM_SB(sb)->super_start,
			(int) b1581->t, (int) b1581->s,
			(int) b1581->disk_format, (int) 0/*b1581->flag*/);
#endif
		cbm_brelse(bh);		/* finished with block */
	}

	/*
	 *	This must be done after the brelse because the bh is
	 *	a dummy allocated by cbm_bread (see buffer.c) [why?-DF]
	 */
	sb->s_blocksize_bits = 9;	/* 512 byte blocks--is this supposed to be the
								   size of the buffer cache blocks? */
	if (error || debug) {
		printk("[CBM FS Rel. " VERSION ",check=%c,conv=%c,"
		    "uid=%d,gid=%d,umask=%03o]\n",check,
		    conversion,uid,gid,umask);
		printk("[sb=%lu,sl=%u,rs=%lu,re=%lu,tb=%lu]\n",
		    CBM_SB(sb)->super_start, CBM_SB(sb)->super_length,
		    CBM_SB(sb)->root_start, CBM_SB(sb)->root_high, CBM_SB(sb)->blocks);
	}
	if (error) {
		if (!silent)
			printk("VFS: Can't find a valid CBM filesystem on dev 0x%04x.\n",
				   sb->s_dev);
		unlock_super(sb);
		sb->s_dev = 0;
		MOD_DEC_USE_COUNT;
		return NULL;
	}
	sb->s_magic = CBM_SUPER_MAGIC;
	CBM_SB(sb)->name_check = check;
	CBM_SB(sb)->conversion = conversion;
	unlock_super(sb);

	/*
	 * Set up enough so that it can read an inode
	 */
	sb->s_op = &cbm_sops;
	CBM_SB(sb)->fs_uid = uid;
	CBM_SB(sb)->fs_gid = gid;
	CBM_SB(sb)->fs_umask = umask;
	CBM_SB(sb)->quiet = quiet;
	CBM_SB(sb)->fastsize = fastsize;
	CBM_SB(sb)->nous = nous;
	CBM_SB(sb)->bam_wait = NULL;
	CBM_SB(sb)->bam_lock = 0;
	CBM_SB(sb)->prev_free = 0;
	CBM_SB(sb)->free_blocks = -1; /* don't know yet */
	if (!(sb->s_mounted = iget(sb, CBM_ROOT_INODE(sb)))) {
		sb->s_dev = 0;
		printk("get root inode failed\n");
		MOD_DEC_USE_COUNT;
		return NULL;
	}
	return sb;
}


/*
 * Return disk free space and free inodes given a superblock
 */

#if LINUX_VERSION_CODE >= 0x010300	/* kernel ver. >= 1.3.0 */
void cbm_statfs(struct super_block *sb,struct statfs *buf, int bufsiz)
#else
#define bufsiz 40
void cbm_statfs(struct super_block *sb,struct statfs *buf)
#endif
{
	int free;

	if (bufsiz < 40)
		return;			/* aborting is easier than changing this routine to
						   return only what is asked */

	put_fs_long(sb->s_magic,&buf->f_type);
	put_fs_long(CBM_LBLOCK_SIZE,&buf->f_bsize); /* should be CBM_DATA_SIZE */
	put_fs_long(CBM_SB(sb)->blocks,&buf->f_blocks);

	/*
	 * Don't change the BAM while we're counting
	 */

	lock_bam(sb);

	/*
	 * Has free_blocks already been determined?
	 */

	if (CBM_SB(sb)->free_blocks >= 0)
		free = CBM_SB(sb)->free_blocks;
	else {
		free = cbm_free_blocks(sb);
		CBM_SB(sb)->free_blocks = free;
	}
	unlock_bam(sb);

	put_fs_long(free,&buf->f_bfree);
	put_fs_long(free,&buf->f_bavail);
	put_fs_long(0,&buf->f_files);		/* don't support inode counting */
	put_fs_long(0,&buf->f_ffree);		/* don't support inode counting */
	put_fs_long(16,&buf->f_namelen);
}

/*
 * Return the disk block given an inode and logical file block
 */

int cbm_bmap(struct inode *inode,int block)
{
	/*
	 * NOTE: There seems to be a problem here.  Returning 0 seems to indicate
	 * an error, but on CBM disks, sector 0 is just as valid as any.
	 * -1 as error would work, but how does the calling function handle that?
	 */ 
	if ((block = get_sector(inode,block)) < 0) return 0 /* try -1 */ ;
	return block;
}

/*
 * Fill in the inode structure at the pointer, given the inode number
 * in inode->i_ino 
 * Much of the directory stuff is 1581-specific and would need to be modified
 * in case directories were desired on 1541 disks.
 */

void cbm_read_inode(struct inode *inode)
{
	struct super_block *sb = inode->i_sb;
	struct buffer_head *bh;
	struct cbm_dir_entry *raw_entry;
	unsigned long blocks;

#if DEBUG
printk("cbm_read_inode: %ld\n",inode->i_ino);
#endif
	CBM_I(inode)->i_busy = 0;
	inode->i_uid = CBM_SB(sb)->fs_uid;
	inode->i_gid = CBM_SB(sb)->fs_gid;

	/*
 	 * Root inode (special case inode)
 	 */
	if (inode->i_ino == CBM_ROOT_INODE(sb)) {
		inode->i_mode = S_IFDIR | (S_IRWXUGO & ~CBM_SB(sb)->fs_umask);
		inode->i_op = &cbm_dir_inode_operations;

		CBM_I(inode)->i_start = CBM_SB(sb)->root_start;
		CBM_I(inode)->i_type = CBM_CBM | CBM_CLOSED;
		CBM_I(inode)->i_text = 0;					/* directory is binary */
		CBM_I(inode)->i_rec_size = 0;
		CBM_I(inode)->i_side_sec = 0;
		CBM_I(inode)->i_parent = inode->i_ino;		/* not really parent,
														but doesn't matter */
		if (CBM_SB(sb)->fastsize)

			/*
			 * Return the size of a standard disk directory
			 */

			blocks = (CBM_SB(sb)->fs_type == 1541 ?
					CBM1541_DIR_LEN : CBM1581_DIR_LEN);
		else {

			/*
			 * Count the actual number of directory blocks here, just in case
			 * the directory is a nonstandard extended one.
			 *
			 * Round up to the least number of blocks needed to hold the data.
			 * This will give strange results if the directory length == 0 
			 * (which should never happen), or if an error occurred while
			 * reading the disk (which might happen)
			 */

			blocks = 1+((chain_len(inode)-1)/CBM_DATA_SIZE);
		}

		inode->i_size = CBM_DATA_SIZE * blocks;
		inode->i_blksize = CBM_LBLOCK_SIZE; /* should be CBM_DATA_SIZE */
		inode->i_blocks = blocks;

		/*
		 * Currently, all subdirectories below root have '..' pointing to
		 * the root, so this next line isn't always technically correct
		 */

		inode->i_nlink = cbm_subdirs(inode)+2;

		/*
		 * CBM doesn't have file times
		 */

		inode->i_mtime = inode->i_atime = inode->i_ctime = 0;

		return;		/* Return with changed root inode information */
	}

	/*
 	 * Inode is from a disk entry
	 *
	 * Convert the inode number into a corresponding disk directory block
	 */

	if (!(bh = cbm_bread(sb,CBM_INOBLOCK(inode->i_ino)))) {
		printk("dev = 0x%04X, ino = %ld\n",inode->i_dev,inode->i_ino);
		panic("cbm_read_inode: unable to read i-node block");
	}
	raw_entry = &((struct cbm_dir_entry *)
		((struct cbm_data_sector *) bh->b_data)->data)[CBM_INODIRENT(inode->i_ino)];
#if DEBUG
printk("cbm_read_inode: name \"%-16s\"\n", raw_entry->name);
#endif

	if (IS_CBM(raw_entry->type)) {

	/*
 	 * Partition or directory entry (1581 only)
 	 */
		struct buffer_head *new_bh;
		struct cbm1581_header_sector *head;

		/*
		 * Must read the first sector of the partition to find what it is
		 */

		if (!(new_bh = cbm_bread(sb, cbm_block(sb, raw_entry->t, raw_entry->s)))) {
			printk("dev = 0x%04X, ino = %ld\n",inode->i_dev,inode->i_ino);
			panic("cbm_read_inode: unable to read i-node block");
		}
		head = (struct cbm1581_header_sector *) new_bh->b_data;
		if (is_1581_header(head)) {
		
			/*
 			 * Partition is a subdirectory
 			 */

			inode->i_mode = CBM_MKMODE(raw_entry->type,S_IRWXUGO &
			    ~CBM_SB(sb)->fs_umask) | S_IFDIR;
			inode->i_op = &cbm_dir_inode_operations;
			inode->i_nlink = cbm_subdirs(inode) + 2;

			CBM_I(inode)->i_start = cbm_block(sb, head->t, head->s);

			if (CBM_SB(sb)->fastsize)
				/*
				 * Return the size of a standard disk directory
				 *
				 * If dirs are to be supported on 1541, then CBM1541_DIR_LEN
				 * isn't correct for some tracks (not enough sectors) but
				 * shouldn't matter
				 */

				blocks = (CBM_SB(sb)->fs_type == 1541 ?
						CBM1541_DIR_LEN : CBM1581_DIR_LEN);

			else {

				/*
				 * Count the actual number of directory blocks here, just in
				 * case the directory is a nonstandard extended one
				 *
				 * Round up to the least number of blocks needed to hold the data
				 * This will give strange results if the directory length == 0 
				 * (which should never happen), or if an error occurred while
				 * reading the disk (which might happen)
				 */

				blocks = 1+((chain_len(inode)-1)/CBM_DATA_SIZE);
			}
			inode->i_blocks = blocks;
			inode->i_size = blocks * CBM_DATA_SIZE;

		} else {

			/*
 			 * Partition entry which isn't a directory (treat as a nonreadable file)
 			 */

			inode->i_mode = S_IFREG;	/* file is not accessible (0 mask) because
										   the partition may not have T&S links */
			inode->i_nlink = 1;			/* CBM doesn't have hard links */
			inode->i_op = &cbm_file_inode_operations;
			inode->i_size = CF_LE_W(raw_entry->blocks) * CBM_DATA_SIZE;
			inode->i_blocks = CF_LE_W(raw_entry->blocks);
			CBM_I(inode)->i_start = cbm_block(sb,raw_entry->t, raw_entry->s);
		}

		cbm_brelse(new_bh);

	} else {

		/*
 		* Regular file inode
 		*/

		inode->i_mode = S_IFREG | CBM_MKMODE(raw_entry->type,
			((IS_NOEXEC(inode) ? S_IRUGO|S_IWUGO : S_IRWXUGO) &
			~CBM_SB(sb)->fs_umask));
		inode->i_op = &cbm_file_inode_operations;
		inode->i_nlink = 1;			/* CBM doesn't have links */
		CBM_I(inode)->i_start = cbm_block(sb,raw_entry->t, raw_entry->s);

		if (CBM_SB(sb)->fastsize)		/* approximate file size */
			inode->i_size = CF_LE_W(raw_entry->blocks) * CBM_DATA_SIZE;
		else
			/*
			 * Count the actual number of bytes here
			 *
			 * BUG: An error while reading the file is ignored!
			 */
			inode->i_size = chain_len(inode);

		inode->i_blocks = CF_LE_W(raw_entry->blocks);
			/* this is for CBM_DATA_SIZE-sized blocks */
	}

	/*
	 * This is common to directories, partitions and files
	 */

	CBM_I(inode)->i_side_sec = cbm_block(sb,raw_entry->ss_t, raw_entry->ss_s);
	CBM_I(inode)->i_rec_size = (unsigned long) raw_entry->rec_size;
	CBM_I(inode)->i_type = raw_entry->type;
	CBM_I(inode)->i_text = is_text(CBM_SB(sb)->conversion,
	    CBM_I(inode)->i_type);

	/*
	 * The parent won't be correct for subdirectories > 1 level down from root
	 */
	CBM_I(inode)->i_parent = CBM_ROOT_INODE(sb);

	inode->i_blksize = CBM_LBLOCK_SIZE;	/* should be CBM_DATA_SIZE */
	inode->i_mtime = inode->i_atime = inode->i_ctime = 0;

	cbm_brelse(bh);
}


#if MAY_NEED_FOR_FUTURE_WRITE_SUPPORT
void msdos_write_inode(struct inode *inode)
{
	struct super_block *sb = inode->i_sb;
	struct buffer_head *bh;
	struct msdos_dir_entry *raw_entry;

	inode->i_dirt = 0;
	if (inode->i_ino == MSDOS_ROOT_INO || !inode->i_nlink) return;
	if (!(bh = bread(inode->i_dev,inode->i_ino >> MSDOS_DPB_BITS,
	    SECTOR_SIZE))) {
		printk("dev = 0x%04X, ino = %ld\n",inode->i_dev,inode->i_ino);
		panic("msdos_write_inode: unable to read i-node block");
	}
	raw_entry = &((struct msdos_dir_entry *) (bh->b_data))
	    [inode->i_ino & (MSDOS_DPB-1)];

	if (S_ISDIR(inode->i_mode)) {
		raw_entry->attr = ATTR_DIR;
		raw_entry->size = 0;
	}
	else {
		raw_entry->attr = ATTR_NONE;
		raw_entry->size = CT_LE_L(inode->i_size);
	}
	raw_entry->attr |= MSDOS_MKATTR(inode->i_mode) |
	    MSDOS_I(inode)->i_attrs;
	raw_entry->start = CT_LE_L(MSDOS_I(inode)->i_start);
	date_unix2dos(inode->i_mtime,&raw_entry->time,&raw_entry->date);
	raw_entry->time = CT_LE_W(raw_entry->time);
	raw_entry->date = CT_LE_W(raw_entry->date);
	mark_buffer_dirty(bh, 1);
	brelse(bh);
}
#endif


/*
 * Verify & process changes to an inode's attributes?
 */

int cbm_notify_change(struct inode * inode,struct iattr * attr)
{
	int error;

	error = inode_change_ok(inode, attr);
	if (error)
		return error;

	if (((attr->ia_valid & ATTR_UID) && 
	     (attr->ia_uid != CBM_SB(inode->i_sb)->fs_uid)) ||
	    ((attr->ia_valid & ATTR_GID) && 
	     (attr->ia_gid != CBM_SB(inode->i_sb)->fs_gid)) ||
	    ((attr->ia_valid & ATTR_MODE) &&
	     (attr->ia_mode & ~CBM_VALID_MODE)))
		error = -EPERM;

	/*
	 * Don't report an illegal mode error in quiet mode
	 */

	if (error)
		return CBM_SB(inode->i_sb)->quiet ? 0 : error;

	inode_setattr(inode, attr);

	if (IS_NOEXEC(inode) && !S_ISDIR(inode->i_mode))
		inode->i_mode &= S_IFMT | S_IRUGO | S_IWUGO;
	else
		inode->i_mode |= S_IXUGO;

	inode->i_mode = ((inode->i_mode & S_IFMT) | ((((inode->i_mode & S_IRWXU
	    & ~CBM_SB(inode->i_sb)->fs_umask) | S_IRUSR) >> 6)*S_IXUGO)) &
	    ~CBM_SB(inode->i_sb)->fs_umask;
	return 0;
}


/*
 * Modules code and kernel interface
 */

#ifdef __cplusplus
extern "C" {
#endif
 
#ifndef MODULE

#error Code to put cbm into the kernel hasn\'t been added--compile as a module

#else

/*
 * Newer kernels already define kernel_version in module.h
 * Unfortunately, I don't know what kernel version started doing that
 * so I'm guessing at 1.3.99
 */
#if LINUX_VERSION_CODE < 0x010399	/* kernel ver. >= 1.3.99 */
char kernel_version[] = UTS_RELEASE;
#endif

static struct file_system_type cbm_fs_type = {
	cbm_read_super, "cbm", 1, NULL
};

int init_module(void)
{
	struct super_block *dummy;

	/*
	 * make sure the CBM superblock extensions aren't too big to fit
	 * in the available space (there's no compile-time way of doing
	 * this that I know of in GCC)
	 */

	if (sizeof(struct cbm_sb_info) > sizeof(dummy->u))
		return -E2BIG;

	register_filesystem(&cbm_fs_type);
	return 0;
}

void cleanup_module(void)
{
	unregister_filesystem(&cbm_fs_type);
}

#endif

#ifdef __cplusplus
}
#endif
