#ifndef _LINUX_CBM_FS_H
#define _LINUX_CBM_FS_H

/*
 * Commodore 1541/1581 filesystem constants/structures
 */
#include <linux/fs.h>
#include <linux/stat.h>
#include <linux/ioctl.h>
#include <linux/dirent.h>
#include <linux/version.h>
#include <endian.h>

#define CBM_SUPER_MAGIC 0x1581	/* Linux filesystem superblock magic number */

#define CBM_BLOCK_SIZE 512	/* physical sector size on disk */
#define CBM_LBLOCK_SIZE 256	/* logical sector size */
#define CBM_DATA_SIZE 254	/* size in bytes of data portion of sector */
#define CBM_DPS	8		/* dir entries per CBM block */
#define CBM_DPB	(CBM1581_DPS) 	/* dir entries per CBM block */
#define CBM_DPB_BITS 3		/* bits needed to hold CBM_DPB entries */
#define CBM_DPB_MASK 7		/* 2^CBM_DPB_BITS - 1 */

#define MAP_CACHE 26 /*259*/	/* entries in file location cache (20 bytes ea.)
				   ought to be big enough to hold a 64k file 
				   for efficient random file access */

#define CBM_PREFETCH 32		/* size of prefetch buffer for file reading */
							/* must be an even number >= 2 */

/* VIC-1541-specific constants */
#define CBM1541_SUPER_LOC cbm1541_block(CBM1541_DIR_TRACK,0)
	/* start of superblock */
#define CBM1541_SUPER_LEN 1		/* length of BAM in CBM blocks */
#define CBM1541_DIR_LEN 18		/* length of root directory in CBM blocks */
#define CBM1541_BLOCKS 683		/* number of CBM blocks in 1541 */
#define CBM1541_TRACKS 35		/* number of tracks in standard 1541 disk */
#define CBM1541_DIR_TRACK 18	/* directory track */
#define CBM1541_DOS_FORMAT 'A'		/* magic number */

/* 1581-specific constants */
#define CBM1581_SUPER_LOC cbm1581_block(CBM1581_DIR_TRACK,0)
	/* start of superblock */
#define CBM1581_SUPER_LEN 3		/* length of superblock in CBM blocks */
#define CBM1581_DIR_LEN 37		/* length of root directory in CBM blocks */
#define CBM1581_BLOCKS 3200		/* number of CBM blocks in 1581 */
#define CBM1581_TRACKS 80		/* number of tracks in standard 1581 disk */
#define CBM1581_DIR_TRACK 40	/* directory track */
#define CBM1581_DOS_FORMAT 'D' /* magic number */

/* File type mask bits (or'ed with cbm_file_types) */
enum cbm_file_attr {
    CBM_LOCKED = 0x40,     /* file locked */
    CBM_CLOSED = 0x80      /* file closed bit */
};

/* File types */
enum cbm_file_types {
    CBM_DEL = 0,
    CBM_SEQ,
    CBM_PRG,
    CBM_USR,
    CBM_REL,
    CBM_CBM,		/* 1581/C65 only */
    CBM_DJJ,		/* C65 only */
    CBM_FAB,		/* C65 only */

    CBM_TYPE_MASK = 0x07        /* Mask to extract file type */
};

#define CBM_NAME 16 			/* maximum name length */
#define CBM_END_NAME 0xA0		/* indicates end of file name */

/* ioctl commands */

#define CBM_IOC_GETFTYPE		_IOR(0x15, 1, long)	/* get file type */
#define CBM_IOC_SETFTYPE		_IOW(0x15, 2, long)	/* set file type */
#define CBM_IOC_GETMEDIA		_IOR(0x15, 3, long)	/* get disk type */
#define CBM_IOC_GETRECSIZE		_IOR(0x15, 4, long)	/* get REL record size */

/* access macros */

#define IS_FREE(n) ((n) == 0)	/* cbm_dir_entry.type shows entry is free to use */
#define IS_REL(t) (((t) & CBM_TYPE_MASK) == CBM_REL)
#define IS_SEQ(t) (((t) & CBM_TYPE_MASK) == CBM_SEQ)
#define IS_CBM(t) (((t) & CBM_TYPE_MASK) == CBM_CBM)

#define CBM_VALID_MODE (S_IFREG | S_IFDIR | S_IRWXU | S_IRWXG | S_IRWXO)
	/* valid file mode bits--should S_IFDIR be included? */

#ifndef MODULE
#define CBM_SB(s) (&((s)->u.cbm_sb))
#define CBM_I(i) (&((i)->u.cbm_i))
#else
/* Use this if the kernel hasn't been patched to add cbm_sb_info and
cbm_inode_info to struct super_block and struct inode, respectively.
Make sure the structs are small enough to fit into the unions! */
#define CBM_SB(s) ((struct cbm_sb_info *)&((s)->u.generic_sbp))
#define CBM_I(i) ((struct cbm_inode_info *)&((i)->u.generic_ip))
#endif

/*
 * Conversion from and to little-endian byte order. (no-op on i386/i486)
 *
 * Naming: Ca_b_c, where a: F = from, T = to, b: LE = little-endian, BE = big-
 * endian, c: W = word (16 bits), L = longword (32 bits)
 */

#if (__BYTE_ORDER != __LITTLE_ENDIAN)
/* big-endian */
#error Test these big-endian conversion macros and anything else to get the filesystem running!
#define CF_LE_W(n) ((((n) & 0xff) << 8) | (((n) & 0xff00) >> 8))
#define CF_LE_L(n) ((((n) & 0xff) << 24) | (((n) & 0xff00) << 8) | \
                    (((n) & 0xff0000L) >> 8) | (((n) & 0xff000000L) >> 24))
/* Haven't bothered with CT_* yet because they're only needed for writing */
#else
/* little-endian */
#define CF_LE_W(v) (v)
#define CF_LE_L(v) (v)
#define CT_LE_W(v) (v)
#define CT_LE_L(v) (v)
#endif

/* File location cache; stores location on disk for blocks in files */
struct map_cache {
	int device;			/* device number. 0 means unused. */
	ino_t ino;			/* inode number. */
	int file_block;		/* block number in the file. */
	int disk_sector;	/* logical sector on device. */
	struct map_cache *next; /* next cache entry */
};

/* The following structures must all be packed onto byte boundaries.  This is
how they happen to end up in i386 GCC 2.5.8.  Others architectures may need
to use __attribute__ ((packed)) on all entries */

/* Contents of one 1541 BAM entry */
struct cbm1541_bam_entry {
	__u8 free_blocks;			/* free blocks in this track */
	__u8 bit_map[3];			/* free block bit map */
};

/* Contents of dir header sector */
struct cbm1541_header_sector {
	__u8 t, s;					/* track & sector of directory */
	__u8 disk_format;			/* 'A' for 4040 disk format */
	__u8 flag;					/* null flag for future use */
	struct cbm1541_bam_entry bam[CBM1541_TRACKS];	/* block availability map */
	__u8 disk_name[16];			/* disk name */
	__u8 space1[2];				/* shifted spaces */
	__u8 disk_id[2];			/* disk id code */
	__u8 space2;				/* shifted space */
	__u8 dos_version[2];		/* '2A' for 1541 */
	__u8 space3[4];				/* shifted spaces */
	__u8 filler[9];				/* unused */
	__u8 string[12];			/* 1541 manual says this sometimes used */
};

/* Contents of one 1581 BAM entry */
struct cbm1581_bam_entry {
	__u8 free_blocks;			/* free blocks in this track */
	__u8 bit_map[5];			/* free block bit map */
};

/* Contents of each of the 2 1581 BAM sectors */
struct cbm1581_bam_sector {
	__u8 t, s;					/* track & sector of next BAM */
	__u8 disk_format;			/* 'D' for 1581 disk format */
	__u8 disk_format_cpl;		/* complemented disk_format value */
	__u8 disk_id[2];			/* disk id code */
	__u8 io_flag;				/* bit 7: verify; bit 6: check header CRC */
	__u8 auto_loader_flag;		/* 0 or ??? */
	__u8 reserved[8];
	struct cbm1581_bam_entry bam[CBM1541_TRACKS];	/* 40 tracks worth of BAM entries */
};

struct cbm1581_header_sector {
	__u8 t, s;				/* track & sector of directory */
	__u8 disk_format;		/* 'D' for 1581 disk format */
	__u8 flag;				/* null flag for future use */
	__u8 disk_name[16];		/* disk name */
	__u8 space1[2];			/* shifted spaces */
	__u8 disk_id[2];		/* disk id code */
	__u8 space2;			/* shifted space */
	__u8 dos_version[2];	/* '3D' for 1581 */
	__u8 space3[2];			/* shifted spaces */
};

/* CBM data sector */
struct cbm_data_sector {
	__u8 t, s;					/* next track & sector of file */
	__u8 data[CBM_DATA_SIZE];	/* file data */
};
	

/* Contents of single directory entry */
struct cbm_dir_entry {
	__u8 type;				/* file type & attr */
	__u8 t, s;				/* first track & sector of file */
	__u8 name[16];			/* file name */
	__u8 ss_t, ss_s;		/* first side sector track & sector */
	__u8 rec_size;			/* relative file record size */
	__u8 unused[4];
	__u8 open_t, open_s;	/* t & s of replacement file for OPEN@ */
	__u16 blocks;			/* size of file (in blocks) */
	__u8 padding[2];		/* doesn't exist for final entry in CBM block */
};

/* Convert CBM attribute bits and a mask to the UNIX mode. */
#define CBM_MKMODE(a,m) ((m) & ((((a) & CBM_LOCKED) ? S_IXUGO : S_IWUGO|S_IXUGO)\
				 |(((a) & CBM_CLOSED) ? S_IRUGO : 0)))

/* Convert the UNIX mode to the CBM attribute bits. */
#define CBM_MKATTR(m) (((m) & S_IWUGO) ? CBM_CLOSED : CBM_LOCKED|CBM_CLOSED)

/* Convert a CBM block and directory entry number to inode number */
#define CBM_MKINODE(b,e) (((b) << CBM_DPB_BITS) | (e))

/* Convert an inode number to CBM block */
#define CBM_INOBLOCK(i) ((i) >> CBM_DPB_BITS)

/* Convert an inode number to directory entry number */
#define CBM_INODIRENT(i) ((i) & CBM_DPB_MASK)

/* Convert a directory file offset into a directory entry number */
#define CBM_OFS_DIRENT(o) (((o) % CBM_DATA_SIZE) / (CBM_LBLOCK_SIZE / CBM_DPS))

/* Get the root inode number for a superblock */
#define CBM_ROOT_INODE(sb) (CBM_MKINODE(CBM_SB(sb)->super_start,0))

#ifdef __KERNEL__

/* These prototypes aren't all converted from the MS-DOS filesystem code yet */

/* misc.c */

extern long chain_len(struct inode *inode);
extern unsigned long cbm_block(struct super_block *sb, int track, int sector);
extern unsigned long cbm1541_block(int track, int sector);
extern void cbm1541_ts(unsigned long block,int *track, int *sector);
extern unsigned long cbm1581_block(int track, int sector);
extern void cbm1581_ts(unsigned long block,int *track, int *sector);
extern long cbm_free_blocks(struct super_block *sb);
extern int is_1541_header(struct cbm1541_header_sector *header);
extern int is_1581_header(struct cbm1581_header_sector *header);
extern void fs_panic(struct super_block *s,char *msg);
extern void lock_creation(void);
extern void unlock_creation(void);
extern void lock_bam(struct super_block *sb);
extern void unlock_bam(struct super_block *sb);
extern int msdos_add_cluster(struct inode *inode);
extern int cbm_get_entry(struct inode *dir,loff_t *pos,struct buffer_head **bh,
    struct cbm_dir_entry **de);
extern int cbm_scan(struct inode *dir,char *name,struct buffer_head **res_bh,
    struct cbm_dir_entry **res_de,int *ino);
extern int msdos_parent_ino(struct inode *dir,int locked);

extern int is_text(char conversion,unsigned char type);
extern int cbm_subdirs(struct inode *dir);

/* map.c */

extern int map_next(struct super_block *sb,int nr,int *size);
extern int bam_access(struct super_block *sb,int nr,int new_value);
extern int cbm_smap(struct inode *inode,int sector);
extern int fat_free(struct inode *inode,int skip);
extern void cache_init(void);
void cache_lookup(struct inode *inode,int block,int *file_blk,int *disk_sec);
void cache_add(struct inode *inode,int f_clu,int d_clu);
void cache_inval_inode(struct inode *inode);
void cache_inval_dev(int device);
int get_sector(struct inode *inode,int sector);

/* namei.c */

extern int cbm_lookup(struct inode *dir,const char *name,int len,
	struct inode **result);
extern int msdos_create(struct inode *dir,const char *name,int len,int mode,
	struct inode **result);
extern int msdos_mkdir(struct inode *dir,const char *name,int len,int mode);
extern int msdos_rmdir(struct inode *dir,const char *name,int len);
extern int msdos_unlink(struct inode *dir,const char *name,int len);
extern int msdos_unlink_umsdos(struct inode *dir,const char *name,int len);
extern int msdos_rename(struct inode *old_dir,const char *old_name,int old_len,
	struct inode *new_dir,const char *new_name,int new_len);

/* inode.c */

extern void cbm_put_inode(struct inode *inode);
extern void cbm_put_super(struct super_block *sb);
extern struct super_block *cbm_read_super(struct super_block *s,
					    void *data,int);
extern int cbm_bmap(struct inode *inode,int block);
extern void cbm_read_inode(struct inode *inode);
extern void cbm_write_inode(struct inode *inode);
extern int cbm_notify_change(struct inode *,struct iattr *);
#if LINUX_VERSION_CODE >= 0x010300	/* kernel ver. >= 1.3.0 */
extern void cbm_statfs(struct super_block *sb,struct statfs *buf, int unknown);
#else
extern void cbm_statfs(struct super_block *sb,struct statfs *buf);
#endif

/* dir.c */

extern struct inode_operations cbm_dir_inode_operations;

#if LINUX_VERSION_CODE >= 0x010300	/* kernel ver. >= 1.3.0 */
extern int cbm_readdir(struct inode *inode, struct file *filp, void *unknown,
				filldir_t kfilldir);
#else
/* dirent in in user space */
extern int cbm_readdir(struct inode *inode, struct file *filp, struct dirent *dirent,
				int count);
#endif

/* file.c */

extern struct inode_operations cbm_file_inode_operations;
extern int cbm_file_read(struct inode *, struct file *, char *, int);
/*extern int cbm_file_write(struct inode *, struct file *, char *, int);*/
extern struct inode_operations cbm_file_inode_operations_no_bmap;
extern void msdos_truncate(struct inode *inode);

/* mmap.c */
extern int msdos_mmap(struct inode *, struct file *, struct vm_area_struct *);

/* ioctl.c */
extern int cbm_ioctl (struct inode * inode, struct file * filp, unsigned int cmd,
	unsigned long arg);

#endif /* __KERNEL__ */

#endif
