/*
 *  linux/fs/cbm/map.c
 *
 *  Written 1995 by Dan Fandrich <dan@fch.wimsey.bc.ca>
 *  Modified from MS-DOS filesystem written 1992,1993 by Werner Almesberger
 */

/*
The map cache holds the location on disk for logical blocks in files.
This means that the file links need not be examined sequentially each
time to find a file's location on disk, especially in a random-access read.
A cache is especially important in the Commodore filesystem since finding a
file's location on disk is O(n) on average otherwise.  Actually, the cache
search algorithm is currently also O(n), but doesn't require disk reads so
is still much faster.
A single cache linked list holds entries for all mounted filesystems.
*/

#include <linux/cbm_fs.h>
#include <linux/cbm_fs_sb.h>
#include <linux/cbm_fs_i.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/stat.h>
#include <limits.h>

#include "cbmbuffer.h"

#define DEBUG 0 /*1*/

#if DEBUG
#define PRINTK(x) printk x	/* debugging */
#else
#define PRINTK(x)
#endif

/* Chain mapping cache for all mounted devices */
static struct map_cache *map_cache,cache[MAP_CACHE];


/* Access 1541 BAM [see bam_access()] */
/* This reads the entire BAM sector each time it is called, so scanning an
   entire disk of 683 blocks will take 683 reads.  Luckily, the BAM ought to
   be in a bread buffer each time if done sequentially so it shouldn't be
   too slow. */
int bam1541_access(struct super_block *sb,int nr,int new_value)
{
/* This is incomplete */
#if 0
	struct buffer_head *bh;
	struct cbm1541_header_sector *sec;
	int track,sector,status;

	/* Determine the BAM offset and bit for the given block */
	cbm1541_ts(nr,&track,&sector);
	
	/* Read the BAM */
	if (!(bh = cbm_bread(sb,CBM_SB(sb)->super_start))) {
		printk("cbm_bread in map_next failed\n");
		return -2;
	}
	sec = (struct cbm1541_header_sector *) bh->b_data;

	status = sec->bam[ofs] & bit;

	if (new_value >=0)
		/* modify entry & mark dirty here */ ;
	cbm_brelse(bh);

	return status;
#endif
	return 1;	/* always show allocated */
}

/* Access 1581 BAM [see bam_access()] */
int bam1581_access(struct super_block *sb,int nr,int new_value) {
	/* NOT YET IMPLEMENTED */

	return 1;	/* always show allocated */
}

/* Returns the status of the BAM block nr;
		0=free, 1=used, -1=range error, -2=I/O error
   If new_value is != -1, that BAM entry is replaced by it. */
int bam_access(struct super_block *sb,int nr,int new_value)
{
	/* sanity check on sector number */
	if (nr >= (long) CBM_SB(sb)->blocks) return -2;

	return (CBM_SB(sb)->fs_type == 1541) ?
				bam1541_access(sb,nr,new_value) :
				bam1581_access(sb,nr,new_value);
}

/*
Reads sector nr and returns the next sector link
If size != NULL, sets it to the number of bytes of data in the supplied sector.
Input: sb, nr, size
Return: sector number or -1 for end of file or -2 for error
		*size
*/
int map_next(struct super_block *sb,int nr,int *size)
{
	struct buffer_head *bh;
	struct cbm_data_sector *sec;
	int next;

	PRINTK(("map_next: %d\n",nr));

	/* sanity check on sector number */
	if (nr >= (long) CBM_SB(sb)->blocks) return -2;

	if (!(bh = cbm_bread(sb,nr))) {
		printk("cbm_bread in map_next failed\n");
		return -2;
	}
	sec = (struct cbm_data_sector *) bh->b_data;

	/* end of file */
	if (sec->t == 0) {
		next = -1;
		if (size) *size = sec->s - 1;

	} else {
		/* return next block */
		next = cbm_block(sb, sec->t, sec->s);
		if (size) *size = CBM_DATA_SIZE;
	}
	cbm_brelse(bh);


	PRINTK(("map_next: next is t%d,s%d =%d (max %ld)\n",(int)sec->t,(int)sec->s,next,CBM_SB(sb)->blocks));

	/* sanity check the link */
	if ((next >= (long) CBM_SB(sb)->blocks) || (next < -1)) return -2;

	return next;
}

/* Build initial cache structures at filesystem initialization time */
void cache_init(void)
{
	static int initialized = 0;
	int count;

	if (initialized) return;
	map_cache = &cache[0];

	/* Build linked list of cache entries */
	for (count = 0; count < MAP_CACHE; count++) {
		cache[count].device = 0;
		cache[count].next = count == MAP_CACHE-1 ? NULL :
		    &cache[count+1];
	}
	initialized = 1;
}


/*
Search cache to retrieve location on disk for a file's sector.
The highest file block already known is passed in, and if the given file
block isn't in the cache, the highest file block in the cache is returned
(both in file_blk).  If block < 0, then the highest file block in the cache
for the given file is returned.

The cache is searched using a linear time search, which is fine for a small
cache, but not great for a larger one.  Since the Commodore cache size is
a few hundred entries, this should eventually be changed.

Input: inode, block, file_blk
Output: disk_sec, file_blk
*/
void cache_lookup(struct inode *inode,int block,int *file_blk,int *disk_sec)
{
	struct map_cache *walk;

	PRINTK(("cache lookup: <0x%04x,%ld> %d (%d,%d) -> ",inode->i_dev,inode->i_ino,block,*file_blk,*disk_sec));

	/* Search for last entry */
	if (block < 0) block = INT_MAX;

	/* Search cache for an entry <= block */
	for (walk = map_cache; walk; walk = walk->next)
		if (inode->i_dev == walk->device && inode->i_ino == walk->ino &&
		    walk->file_block <= block && walk->file_block > *file_blk) {
			*disk_sec = walk->disk_sector;
			PRINTK(("cache hit: %d (%d)\n",walk->file_block,*disk_sec));
			if ((*file_blk = walk->file_block) == block) return;
		}
	PRINTK(("cache miss\n"));
}

#if DEBUG
/* Display the entire contents of the map cache (for debugging) */
static void list_cache(void)
{
	struct map_cache *walk;

	for (walk = map_cache; walk; walk = walk->next) {
		if (walk->device)
			printk("<0x%04x,%d>(%d,%d) ",walk->device,walk->ino,
			    walk->file_block,walk->disk_sector);
		else printk("-- ");
	}
	printk("\n");
}
#endif


/*
Add an entry to the cache for this inode, file block and disk sector.
The entry is placed at the head of the cache linked list for faster access
next time.
*/
void cache_add(struct inode *inode,int file_blk,int disk_sec)
{
	struct map_cache *walk,*last;

	PRINTK(("cache add: <%d,%ld> %d (%d)\n",inode->i_dev,inode->i_ino,file_blk,disk_sec));
	last = NULL;
	for (walk = map_cache; walk->next; walk = (last = walk)->next)
		if (walk->ino == inode->i_ino && inode->i_dev == walk->device &&
		    walk->file_block == file_blk) {

			/* block is already in the cache */
			if (walk->disk_sector != disk_sec) {
				printk("map cache corruption");
				cache_inval_inode(inode);
				return;
			}
			/* update LRU */
			if (last == NULL) return;
			last->next = walk->next;
			walk->next = map_cache;
			map_cache = walk;
#if DEBUG
			list_cache();
#endif
			return;
		}

	/* add new block to cache */
	/* walk points to the last item in the linked list */
	walk->device = inode->i_dev;
	walk->ino = inode->i_ino;
	walk->file_block = file_blk;
	walk->disk_sector = disk_sec;

	/* move entry to head of list */
	last->next = NULL;
	walk->next = map_cache;
	map_cache = walk;
#if DEBUG
	list_cache();
#endif
}


/* Cache invalidation occurs rarely, thus the LRU chain is not updated. It
   fixes itself after a while. */

void cache_inval_inode(struct inode *inode)
{
	struct map_cache *walk;

	for (walk = map_cache; walk; walk = walk->next)
		if (walk->device == inode->i_dev && walk->ino == inode->i_ino)
			walk->device = 0;
}

/* Invalidate every entry in the cache for a given device */
void cache_inval_dev(int device)
{
	struct map_cache *walk;

	for (walk = map_cache; walk; walk = walk->next)
		if (walk->device == device) walk->device = 0;
}


/*
Return the disk sector number given an inode and a logical sector number
i.e. map a file's logical sector onto the disk
Uses the mapping cache.
Returns -1 on an undefined mapping, -2 on error (same as map_next)
*/
int get_sector(struct inode *inode,int sector)
{
	int nr,count;

	/* file is empty */
	if ((nr = CBM_I(inode)->i_start) < 0) return -1;

	/* the first sector number is stored in the inode--no lookup needed */
	if (sector == 0) return nr;

	/* see if cache holds the entry we need, otherwise read mappings from
	   the disk, one at a time, until we get to where we want */
	count = 0;
	for (cache_lookup(inode,sector,&count,&nr); count < sector; count++) {
		if ((nr = map_next(inode->i_sb,nr,NULL)) < 0) return nr;

		cache_add(inode,count+1,nr);
	}

	return nr;
}


/*
Return the disk sector number given an inode and a logical sector number
i.e. map a file's logical sector onto the disk
Returns -1 on an undefined mapping
*/
int cbm_smap(struct inode *inode,int sector)
{
	return get_sector(inode,sector);
}

#if MAY_NEED_FOR_FUTURE_WRITE_SUPPORT
/* Free all clusters after the skip'th cluster. Doesn't use the cache,
   because this way we get an additional sanity check. */

int fat_free(struct inode *inode,int skip)
{
	int nr,last;

	if ((nr = CBM_I(inode)->i_start) < 0) return 0;
	last = 0;
	while (skip--) {
		last = nr;
		if ((nr = fat_access(inode->i_sb,nr,-1)) == -1) return 0;
		if (!nr) {
			printk("fat_free: skipped EOF\n");
			return -EIO;
		}
	}
	if (last)
		fat_access(inode->i_sb,last,MSDOS_SB(inode->i_sb)->fat_bits ==
		    12 ? 0xff8 : 0xfff8);
	else {
		CBM_I(inode)->i_start = -1;
		inode->i_dirt = 1;
	}
	lock_fat(inode->i_sb);
	while (nr != -1) {
		if (!(nr = fat_access(inode->i_sb,nr,0))) {
			fs_panic(inode->i_sb,"fat_free: deleting beyond EOF");
			break;
		}
		if (MSDOS_SB(inode->i_sb)->free_clusters != -1)
			MSDOS_SB(inode->i_sb)->free_clusters++;
		inode->i_blocks -= MSDOS_SB(inode->i_sb)->cluster_size;
	}
	unlock_fat(inode->i_sb);
	cache_inval_inode(inode);
	return 0;
}
#endif
