/*
 * lscbm.c		- List file types on a cbm file system
 *
 * Written 1995 by Daniel Fandrich <dan@fch.wimsey.bc.ca>
 *
 * Based on code
 * Copyright (C) 1993, 1994  Remy Card <card@masi.ibp.fr>
 *                           Laboratoire MASI, Institut Blaise Pascal
 *                           Universite Pierre et Marie Curie (Paris VI)
 *
 * This file can be redistributed under the terms of the GNU General
 * Public License
 */

/*
 * History:
 * 95-04-07	- Created
 */

#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <linux/cbm_fs.h>

/*
#include "et/com_err.h"
#include "e2p/e2p.h"

#include "../version.h"
*/

#define com_err(p,e,s,a) fprintf(stderr, "%s: %s " s "\n", p, sys_errlist[e], a)

const char * program_name = "lscbm";

int all = 0;
int dir_opt = 0;
int rec_opt = 0;
int recursive = 0;
int verbose = 0;

static void volatile usage (void)
{
	fprintf (stderr, "Usage: %s [-Radrv] [files...]\n"
					 "List file types on a CBM filesystem\n", program_name);
	exit (1);
}

static void list_media (const char * name)
{
	int type;

	if (fgetmedia (name, &type) == -1)
		com_err (program_name, errno, "While reading media type on %s",
			 name);
	else
		printf("Disk type is %d\n", type);
}

static void list_types (const char * name)
{
	int type;
	unsigned long rec_size;

	if (fgettype (name, &type) == -1)
		com_err (program_name, errno, "While reading type on %s",
			 name);
	else if (IS_REL(type) && fgetrecsize (name, &rec_size) == -1)
		com_err (program_name, errno, "While reading record size on %s",
			 name);
	else
	{
		print_attributes (stdout, type);
		print_types (stdout, type);
		if (rec_opt)
			printf ("%5lu ", IS_REL(type) ? rec_size : 0);
		printf (" %s\n", name);
	}
}

static int lscbm_dir_proc (const char *, struct dirent *, void *);

static void lscbm_args (const char * name)
{
	struct stat st;

	if (lstat (name, &st) == -1)
		com_err (program_name, errno, "while stating %s", name);
	else
	{
		if (verbose)
			list_media(name);
		if (S_ISDIR(st.st_mode) && !dir_opt)
			iterate_on_dir (name, lscbm_dir_proc, (void *) NULL);
		else
			list_types (name);
	}
}

static int lscbm_dir_proc (const char * dir_name, struct dirent * de, void * private)
{
	char path [MAXPATHLEN];
	struct stat st;

	sprintf (path, "%s/%s", dir_name, de->d_name);
	if (lstat (path, &st) == -1)
		perror (path);
	else
	{
		if (de->d_name[0] != '.' || all)
		{
			list_types (path);
			if (S_ISDIR(st.st_mode) && recursive &&
			    strcmp (de->d_name, ".") && strcmp (de->d_name, ".."))
			{
				printf ("\n%s:\n", path);
				iterate_on_dir (path, lscbm_dir_proc, (void *) NULL);
				printf ("\n");
			}
		}
	}
	return 0;
}

void main (int argc, char ** argv)
{
	char c;
	int i;

/*
	fprintf (stderr, "%s %s, %s for CBM FS %s, %s\n",
		program_name,
		 CBMFSPROGS_VERSION, CBMFSPROGS_DATE,
		 CBMFS_VERSION, CBMFS_DATE);
*/
	if (argc && *argv)
		program_name = *argv;
	while ((c = getopt (argc, argv, "Radlvr?h")) != EOF)
		switch (c)
		{
			case 'R':
				recursive = 1;
				break;
			case 'a':
				all = 1;
				break;
			case 'd':
				dir_opt = 1;
				break;
			case 'r':
				rec_opt = 1;
				break;
			case 'v':
				verbose = 1;
				break;
			default:
				usage ();
		}

	if (optind > argc - 1)
		lscbm_args (".");
	else
		for (i = optind; i < argc; i++)
			lscbm_args (argv[i]);
}
